program makenode;
{
Build up a node and arc net for conversion into a map
Each node can connect up to 4 other adjacent nodes
nx,ny,nz define size of grid. Thresh is probability of arc.

Output file format is:

Random_seed
nodes_in_group_i
x1 y1 z1 link_a link_b link_c
..
xn yn zn link_a link_b link_c

where x1,y1 are locations in cartesian co-ordinates
link_xx is the sequential number of the target node or 0 if none


Revision History
1.0  23-Aug-98  3d compatibility and variable groups and link prob
}

Uses Dos;

const version='Random node-arc network maker v1.00';
      Program_max_groups = 8;
      Thresh_max  = 50;
      Thresh_default = 35;
      issue_date='30-Aug-98 ';
      program_name='MAKENODE';
      default_output_file_name = 'NODES';
      default_output_extension = 'TXT';
      author_email='howell@anglianet.co.uk';
      author_name='Jeff Howell ';

Procedure ShowHelp;
{ Gives simple help screen }
begin
   Writeln('Usage: ',program_name,' seed [outfile [groupcount [groupsize]]]');
   Writeln('- seed = 32 bit integer seed for random number generation.');
   Writeln('  Can be any value, if 0 then computer chooses.');
   Writeln('- outfile = Map topology file to be created.');
   Writeln('  File name is optional, default name is ',default_output_file_name,',');
   Writeln('  this file is used by MAKEMAP1.EXE to create the Duke3d map.');
   Writeln('- groupcount = Number of separate node groups.');
   Writeln('  User must specify number from 1 to ',Program_max_groups,'.');
   Writeln('- groupsize = Factor to control size of each group of nodes.');
   Writeln('  An optional number from 1 (small) to ',thresh_max,
           ' (huge), default is ',thresh_default,'.');
   Writeln(author_name,issue_date,author_email);
   halt;
end;

const nx=8; ny=8;
      Program_min_thresh = 0.05;
      Program_max_thresh = 0.3;

Procedure Parse_filename (var s1: string; s: string);
{ add s to s1 if it does not contain dot }
{ 20-Aug-98 improved path analysis }
var i: integer;
begin
   s1:=Fexpand(s1);
   i:=pos('.',s1);
   If i>0 then exit;
   s1:=s1+'.'+s;
end;

Function Unpack_parameter(param, min, max: longint): longint;
{ returns paramater checked withing range min to max as longint }
var x: longint;
    stat: integer;
begin
   Val(paramstr(param),x,stat);
   Unpack_parameter := x;
   if (stat=0) and (x>=min) and (x<=max) then exit;
   Write('error - parameter ',param);
   if stat<>0 then
      Writeln(' is not an integer')
   else
      Writeln(' is not in range ',min,' to ',max);
   halt;
end;

{ define record structures for linked lists }
{ node_type is the master list from which arc_type is derived }

type arc_type = record
                   cost: integer;     { cost of getting to neighbour }
                   arcx,arcy: integer;      { location of neighbour }
                end;

type node_type = record
                   n,e,s,w: boolean;  { true if arc starts/ends here }
                   g,m: integer;      { group, sequential number of node }
                   link: ^arc_type;   { pointer to link arc or nil }
                 end;


var node: array[1..nx,1..ny] of node_type;
    linked: array [1..nx*ny] of boolean;
    groupsize,grouplist: array [1..nx*ny] of integer;
    proximity: array[1..Program_max_groups,1..Program_max_groups] of arc_type;
    max_groups : integer;

Procedure Random_net(thresh: real);
{ define random interconnections with probability of thresh }
var x,y : integer;
begin
   for x:=1 to nx do
     for y:=1 to ny do
        with node[x,y] do begin
           n:=(random<thresh) and (y>1);
           s:=(random<thresh) and (y<8);
           e:=(random<thresh) and (x<8);
           w:=(random<thresh) and (x>1);
           g:=x+ny*(y-1);
           link:=nil;
        end;
end;

Procedure Make_bidirectional;
{ make links bidirectional }
var x,y : integer;
begin
   for x:=1 to nx do
     for y:=1 to ny do
        with node[x,y] do begin
           if n then node[x,y-1].s:=true;
           if s then node[x,y+1].n:=true;
           if e then node[x+1,y].w:=true;
           if w then node[x-1,y].e:=true;
        end;
end;

Procedure Group_nodes_together;
{ identify independent groups }
var x,y: integer;
    swop: boolean;

procedure min(var a,b: integer);
{ sub procedure of Group_nodes_together }
{ make both numbers equal to the smallest one }
{ if they are not equal and set flag 'swop' }
begin
 if a=b then exit;
 swop:=true;
 if a<b then b:=a else a:=b;
end;

begin { Group_nodes_together }
   repeat
    swop:=false;
    for x:=1 to nx do
       for y:=1 to ny do
          with node[x,y] do begin
             if s then min(g,node[x,y+1].g);
             if e then min(g,node[x+1,y].g);
          end;
    until not swop;
end;   { Group_Nodes_Together }

Procedure Find_largest_groups;
{ count number of nodes in each group and sort into ascending order }
var i,j,x,y,temp: integer;
begin
{ count sizes of groups }
   for i:=1 to nx*ny do
     groupsize[i]:=0;
   for x:=1 to nx do
     for y:=1 to ny do
       with node[x,y] do
         inc(groupsize[g]);

{ find largest groups }
   for i:=1 to nx*ny do grouplist[i]:=i;
   for i:=1 to nx*ny do
      for j:=1 to i-1 do
        if groupsize[grouplist[i]]>groupsize[grouplist[j]] then begin
           temp:=grouplist[i];
           grouplist[i]:=grouplist[j];
           grouplist[j]:=temp;
        end;
end;

Procedure Renumber_groups;
{ convert arbitrary group numbers into ascending sequence of size }
var i,j,k: integer;
begin
   for i:=1 to nx do
      for j:=1 to ny do
         with node[i,j] do begin
            k:=1;
            while (grouplist[k]<>g) and (k<=max_groups) do inc(k);
            g:=k;
         end;
end;

Procedure Build_proximity_tables;
{ fill prox_score and prox_arc with distances and links between groups }
var i,j,x1,y1,g1,x2,y2,g2,my_cost: integer;
    prox_ptr: ^arc_type;

Procedure Update_arc(group_a,group_b,x,y: integer);
{ sub_procedure of Build_proximity_tables }
{ check score of this node against best of group and update if less }
begin
   with proximity[group_a,group_b] do
      if my_cost <= cost then begin
         cost:=my_cost;
         arcx:=x;
         arcy:=y;
       end;
end;

begin { Build_proximity_table }
{ clear all scores }
  for i:=1 to max_groups do
    for j:=1 to max_groups do
       proximity[i,j].cost:=1000;

{ search for initiating nodes }
  for x1:=1 to nx do
     for y1:=1 to ny do begin
       g1:=node[x1,y1].g;
       if g1<=max_groups then

{ search for terminating nodes }
          for x2:=1 to nx do
             for y2:=1 to ny do begin
                g2:=node[x2,y2].g;
                if g2<=max_groups then begin
                    my_cost:=abs(x1-x2)+abs(y1-y2);
                    update_arc(g1,g2,x2,y2);
                    update_arc(g2,g1,x1,y1);
                end;
             end;
     end;
end;

Procedure Minimum_cost_sequence;
{ Find lowest cost path through groups }
{ recursive system to allow for any number of groups }
{ re-uses grouplist array which is now redundant }

type sequence_type = record
                        step_: array[1..Program_max_groups] of integer;
                        score_: integer;
                     end;

var seq, best : sequence_type;

Function cost_of_seq: integer;
{ computes cost of linking sequence described in seq }
var i,my_cost: integer;
begin
   my_cost:=0;
   for i:=1 to max_groups-1 do
      with seq do
         with proximity[step_[i],step_[i+1]] do
            inc(my_cost,cost);
   cost_of_seq:=my_cost;
end;

Procedure check_step(from: integer);
{ Calculates minimum cost of reaching last group from given step }
{ if cost is better (lower) than previous best, replace it }
var i: integer;
begin
   if from>max_groups then with seq do begin
      score_:= cost_of_seq;
      if best.score_>score_ then best:=seq;
   end else with seq do
      for i:= 1 to max_groups do
         if step_[i]=0 then begin
            step_[i]:=from;
            check_step(from+1);
            step_[i]:=0;
         end;
end;

var i: integer;

begin
   with best do begin
      for i:=1 to max_groups do step_[i]:=0;
      score_ := 1000;
   end;
   seq:=best;
   check_step(1);
   for i:=1 to max_groups do
      grouplist[i]:=best.step_[i];
end;

Procedure link_nodes;
{ set link flags in nodes which lie at start of link arcs }
var i,g1,g2: integer;
begin
   for i:=1 to max_groups-1 do begin
      g1:= grouplist[i];
      g2:= grouplist[i+1];
      with proximity[g2,g1] do
          node[arcx,arcy].link := @proximity[g1,g2];
   end;
end;

Procedure Number_net;
{ allocate sequential number for each node starting from 0 }
var x,y,i,count: integer;
begin
   count:=0;
   for i:=1 to max_groups do begin
     for x:=1 to nx do
       for y:=1 to ny do
         with node[x,y] do
           if g=i then begin
             m:=count;
             inc(count);
           end;
   end;
end;

Function In_max_groups(g: integer): boolean;
{ returns true if group g is a member of the working set }
var i: integer;
begin
   In_max_groups:=g<=max_groups;
end;

function group_size(target: integer): integer;
var i,j,count: integer;
begin
   count:=0;
   for i:=1 to nx do
     for j:=1 to ny do
       with node[i,j] do
        if (g=target) then
          inc(count);
   group_size:=count;
end;


Procedure Show_net;
{ display net on VDU }
var i,x,y: integer;
begin
{ list groups }
   for i:= 1 to max_groups-1 do
      writeln('group ',i,' has ',group_size(i),' nodes,');
   writeln('group ',max_groups,' has ',group_size(max_groups),' nodes');

{ show map }
    for y:=1 to ny do begin
       for x:=1 to nx do
          with node[x,y] do
             if n then write('  .  ') else write('     ');
       writeln;
       for x:=1 to nx do
          with node[x,y] do begin
             if w then write('.') else write(' ');
             if in_max_groups(g) then
               write(g:3)
             else
               write ('   ');
             if e then write('.') else write(' ');
          end;
       writeln;
    end;
end;

Procedure Save_net(fn: string);
{ output net }
var f: text;

Procedure  Output_group(groupno: integer);
{ subprocedure of Save_net }
var x,y,n_arcs: integer;

procedure put_arc(n: integer);
{ sub_procedure of Output_group }
begin
   Write(f,n,' ');
   inc(n_arcs);
end;

begin
   writeln(f,group_size(groupno));
   for x:=1 to nx do
     for y:=1 to ny do
       with node[x,y] do
         if g=groupno then begin
            Write(f,x,' ',y,' ',0,' ');      { dummy z value }
            n_arcs:=0;
            if link<>nil then
               with link^ do
                  Put_arc(node[arcx,arcy].m);
            if e then Put_arc(node[x+1,y].m);
            if s then Put_arc(node[x,y+1].m);
            while n_arcs<3 do put_arc(0);
            writeln(f);
         end;
end;

var i: integer;

begin  { save_net }
   assign(f,fn);
   rewrite(f);
   Writeln(f,RandSeed);
   for i:=1 to max_groups do Output_group(i);
   close(f);
end;

var i, thresh_factor: integer;
    node_fn: string;

Begin { main program }
   Writeln(Version);

{ load user parameters and filenames }
   If ParamCount=0 then ShowHelp;
   RandSeed:=Unpack_parameter(1,-2147483647,2147483647);
   If RandSeed=0 then begin
      Randomize;
      Writeln('Using seed=',RandSeed);
   end;

   node_fn:=default_output_file_name;
   If Paramcount>1 then node_fn:=ParamStr(2);
   Parse_filename (node_fn, default_output_extension);

   max_groups:=4;
   If paramCount>2 then
      max_groups:=Unpack_parameter(3,1,Program_max_groups);

   thresh_factor:=thresh_default;
   if ParamCount>3 then
     thresh_factor:= Unpack_Parameter(4,1,thresh_max);

{ call procedures to construct network }
   Random_net(Program_min_thresh+thresh_factor/thresh_max*Program_max_thresh);
   Make_bidirectional;
   Group_nodes_together;
   Find_largest_groups;
   Renumber_groups;
   Build_proximity_tables;
   Minimum_cost_sequence;
   Renumber_groups;
   Link_nodes;
   Number_net;
   Show_net;
   Save_net(node_fn);
end.